"""iOS .IPA asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.file.ios.ipa')
class IOSIpa(asset.Asset):
    """IOS .IPA target asset."""
    content: bytes
    path: Optional[str] = None

    def __str__(self):
        return f'iOS({self.path})'



"""Agent asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.agent')
class Agent(asset.Asset):
    """Agent asset."""
    key: str
    version: Optional[str] = None
    git_location: Optional[str] = None
    docker_location: Optional[str] = None
    yaml_file_location: Optional[str] = None

    def __str__(self):
        if self.version is not None:
            return f'Agent {self.key}:{self.version}'
        else:
            return f'Agent {self.key}'



"""IPv6 address asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.ip.v6')
class IPv6(asset.Asset):
    """IPv6 Address target asset."""
    host: str
    version: int = 6
    mask: Optional[str] = None

    def __str__(self):
        return f'{self.host}/{self.mask}'



"""Runtime are in charge of running scan as defines by a set of agents, agent group and a target asset."""
import abc
import dataclasses
from typing import List, Optional

from ostorlab.assets import asset as base_asset
from ostorlab.cli import dumpers
from ostorlab.runtimes import definitions


@dataclasses.dataclass
class Scan:
    """Scan object."""
    id: str
    asset: Optional[str]
    created_time: str
    progress: Optional[str]


class Runtime(abc.ABC):
    """Runtime is in charge of preparing the environment to trigger a scan."""

    @abc.abstractmethod
    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime is capable of running the provided agent run definition.

        Args:
            agent_group_definition: The agent run definition from a set of agents and agent groups.

        Returns:
            True if can run, false otherwise.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def scan(self, title: str, agent_group_definition: definitions.AgentGroupDefinition,
             asset: base_asset.Asset) -> None:
        """Triggers a scan using the provided agent run definition and asset target.

        Args:
            title: Scan title
            agent_group_definition: The agent run definition from a set of agents and agent groups.
            asset: The scan target asset.

        Returns:
            None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def stop(self, scan_id: str) -> None:
        """Stops a scan with the given id.

        Args:
            scan_id: The scan or universe id.

        Returns:
            None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def list(self, page: int = 1, number_elements: int = 10) -> List[Scan]:
        """Lists scans managed by runtime.

        Args:
            page: Page number for list pagination (default 1).
            number_elements: count of elements to show in the listed page (default 10).

        Returns:
            List of scan objects.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def install(self) -> None:
        """Run runtime installation routine.

        Returns:
            None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def dump_vulnz(self, scan_id: int, dumper: dumpers.VulnzDumper):
        """Dump vulnerabilities to a file in a specific format.
            Returns:
            None
        """
        raise NotImplementedError()



"""Docker builder with log progress streaming."""
import itertools
import re
from typing import Generator, Dict

import docker
from docker.errors import BuildError
from docker.models import images
from docker.models.resource import Collection
from docker.utils.json_stream import json_stream


class BuildProgress(Collection):
    """Build log progress streamer."""

    model = images.Image

    def __init__(self):
        super().__init__(client=docker.from_env())

    def build(self, **kwargs) -> Generator[Dict, None, None]:
        """Build command that copies the initial implementation with log streaming.

        Args:
            **kwargs:

        Yields:
            Log dict with stream or aux as keys and log content in the value.
        """
        resp = self.client.api.build(**kwargs)
        if isinstance(resp, str):
            return self.get(resp)
        last_event = None
        image_id = None
        result_stream, internal_stream = itertools.tee(json_stream(resp))
        for chunk in internal_stream:
            yield chunk
            if 'error' in chunk:
                raise BuildError(chunk['error'], result_stream)
            if 'stream' in chunk:
                match = re.search(
                    r'(^Successfully built |sha256:)([0-9a-f]+)$',
                    chunk['stream']
                )
                if match:
                    image_id = match.group(2)
            last_event = chunk
        if image_id:
            return (self.get(image_id), result_stream)
        raise BuildError(last_event or 'Unknown', result_stream)

    def get(self, key):
        """
        Gets an image.

        Args:
            key (str): The name of the image.

        Returns:
            (:py:class:`Image`): The image.

        Raises:
            :py:class:`docker.errors.ImageNotFound`
                If the image does not exist.
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.prepare_model(self.client.api.inspect_image(key))



"""Utils API to play with Docker Volumes."""

import io
import tarfile
import time

from typing import Dict
import docker
from docker import types as docker_types
from docker import errors as docker_errors

from ostorlab.utils import strings

TEMP_IMAGE = 'busybox:latest'
TEMP_DESTINATION = '/dst'


class VolumeWriter:
    """Volume Writer helper class.

    This class uses a temp image to spawn a container with a target volume attached, write content to it using the
    copy API, and then kill everything.
    """

    def __init__(self):
        self._client = docker.from_env()

    def write(self, volume_name: str, contents: Dict[str, bytes]):
        """Write content volume at path.

        Args:
            volume_name: The volume name to create. This command will override existing volumes with the same name.
            contents: Dict of path where the content will written to in the volume and content in bytes format.
        Returns:
            None
        """
        self._install_image()
        self._create_volume(volume_name)
        self._write_content(volume_name, contents)

    def _install_image(self):
        """Installs a busybox image for the form."""
        self._client.images.pull(TEMP_IMAGE)

    def _create_volume(self, name: str):
        """Override existing images."""
        try:
            self._client.volumes.get(name).remove()
        except docker_errors.NotFound:
            pass
        self._client.volumes.create(name)

    def _prepare_tar(self, contents: Dict[str, bytes]):
        """Copy API expects a tar, this api prepares one with the file content."""
        pw_tarstream = io.BytesIO()
        with tarfile.TarFile(fileobj=pw_tarstream, mode='w') as pw_tar:
            for path, content in contents.items():
                tarinfo = tarfile.TarInfo(name=path)
                tarinfo.size = len(content)
                tarinfo.mtime = time.time()
                pw_tar.addfile(tarinfo, io.BytesIO(content))
        pw_tarstream.seek(0)
        return pw_tarstream

    def _write_content(self, volume_name: str, contents: Dict[str, bytes]):
        """Use the docker copy API to write content to target volume."""
        temp_container_name = strings.random_string(9)
        self._client.containers.run(
            TEMP_IMAGE,
            name=temp_container_name,
            command='sleep infinity',
            detach=True,
            remove=True,
            mounts=[docker_types.Mount(target=TEMP_DESTINATION, source=volume_name, type='volume')]
        )
        pw_tarstream = self._prepare_tar(contents)

        container = self._client.containers.get(temp_container_name)
        container.put_archive(TEMP_DESTINATION, pw_tarstream)
        container.stop()


def create_volume(volume_name: str, contents: Dict[str, bytes]) -> None:
    """Poor man's API to create a volume and persist content.

    Docker do not provide an API to write directly to volume, most likely due to the plugin support, even though that
     should not make it impossible.
    This API is big `hack`, that start a temporary container to copy content to it. The destination is the volume mount
    point.

    Args:
        volume_name: The volume name to create. This command will override existing volumes with the same name.
        contents: Dict of path where the content will written to in the volume and content in bytes format.

    Returns:
        None
    """
    VolumeWriter().write(volume_name, contents)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vulnerabilities.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='vulnerabilities.proto',
    package='',
    serialized_pb=_b(
        '\n\x15vulnerabilities.proto\"Z\n\rvulnerability\x12\r\n\x05title\x18\x02 \x02(\t\x12\x18\n\x10technical_detail\x18\x03 \x02(\t\x12\x13\n\x0brisk_rating\x18\x04 \x02(\t\x12\x0b\n\x03\x64na\x18\x05 \x02(\t\"v\n\x0fvulnerabilities\x12\x0f\n\x07scan_id\x18\x01 \x02(\x05\x12\x12\n\nagent_name\x18\x02 \x02(\t\x12\x15\n\ragent_version\x18\x03 \x02(\t\x12\'\n\x0fvulnerabilities\x18\x04 \x03(\x0b\x32\x0e.vulnerability')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_VULNERABILITY = _descriptor.Descriptor(
    name='vulnerability',
    full_name='vulnerability',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='title', full_name='vulnerability.title', index=0,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='technical_detail', full_name='vulnerability.technical_detail', index=1,
            number=3, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='risk_rating', full_name='vulnerability.risk_rating', index=2,
            number=4, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='dna', full_name='vulnerability.dna', index=3,
            number=5, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=25,
    serialized_end=115,
)

_VULNERABILITIES = _descriptor.Descriptor(
    name='vulnerabilities',
    full_name='vulnerabilities',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='scan_id', full_name='vulnerabilities.scan_id', index=0,
            number=1, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='agent_name', full_name='vulnerabilities.agent_name', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='agent_version', full_name='vulnerabilities.agent_version', index=2,
            number=3, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='vulnerabilities', full_name='vulnerabilities.vulnerabilities', index=3,
            number=4, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=117,
    serialized_end=235,
)

_VULNERABILITIES.fields_by_name['vulnerabilities'].message_type = _VULNERABILITY
DESCRIPTOR.message_types_by_name['vulnerability'] = _VULNERABILITY
DESCRIPTOR.message_types_by_name['vulnerabilities'] = _VULNERABILITIES

vulnerability = _reflection.GeneratedProtocolMessageType('vulnerability', (_message.Message,), dict(
    DESCRIPTOR=_VULNERABILITY,
    __module__='vulnerabilities_pb2'
    # @@protoc_insertion_point(class_scope:vulnerability)
))
_sym_db.RegisterMessage(vulnerability)

vulnerabilities = _reflection.GeneratedProtocolMessageType('vulnerabilities', (_message.Message,), dict(
    DESCRIPTOR=_VULNERABILITIES,
    __module__='vulnerabilities_pb2'
    # @@protoc_insertion_point(class_scope:vulnerabilities)
))
_sym_db.RegisterMessage(vulnerabilities)

# @@protoc_insertion_point(module_scope)



"""Create Agent Group via an API Request."""
import json
from typing import Dict, List, Optional, Union

from ostorlab.apis import request


AgentType = Dict[str, Union[str, List]]


class CreateAgentGroupAPIRequest(request.APIRequest):
    """Persist agent group API request."""

    def __init__(self, name: str, description: str, agents: List[AgentType]) -> None:
        """Initializer"""
        self._name = name
        self._description = description
        self._agents = agents

    @property
    def query(self) -> Optional[str]:
        """Sets the query of the API request.

        Returns:
            The query to create the agent group.
        """
        return """
            mutation PublishAgentGroup($agentGroup: AgentGroupCreateInputType!){
                publishAgentGroup(agentGroup: $agentGroup) {
                    agentGroup{
                        id
                    }
                }
            }
        """


    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request.

        Returns:
            The body of the create agent group request.
        """

        variables = {
            'agentGroup': {
                'name': self._name,
                'description': self._description,
                'access': 'PRIVATE',
                'agents': self._agents
            }
        }
        data = {
            'query': self.query,
            'variables': json.dumps(variables)
        }
        return data



"""Agent Build commands."""
import io
import logging

import click
import docker
from docker import errors

from ostorlab.agent.schema import loader
from ostorlab.agent.schema import validator
from ostorlab.cli import console as cli_console
from ostorlab.cli import docker_requirements_checker
from ostorlab.cli.agent import agent
from ostorlab.cli.agent.build import build_progress

console = cli_console.Console()

logger = logging.getLogger(__name__)


def _image_exists(image: str) -> bool:
    """Checks if container images exists."""
    try:
        docker_sdk_client = docker.from_env()
        docker_sdk_client.images.get(image)
        return True
    except docker.errors.ImageNotFound:
        return False


def _build_image(agent_name: str, container_name: str, dockerfile_path: str, docker_build_root: str,
                 agent_definition_file: io.FileIO) -> None:
    """Build agent image from agent settings."""
    console.info(
        f'Building agent [bold red]{agent_name}[/] dockerfile [bold red]{dockerfile_path}[/]'
        f' at root [bold red]{docker_build_root}[/].')
    with console.status(f'Building [bold red]{container_name}[/]'):
        for log in build_progress.BuildProgress().build(path=docker_build_root,
                                                        dockerfile=dockerfile_path,
                                                        tag=container_name,
                                                        labels={'agent_definition': agent_definition_file.read().decode(
                                                            'utf-8')}):

            if 'stream' in log and log['stream'] != '\n':
                console.info(log['stream'][:-1])
            elif 'error' in log:
                console.error(log['error'][:-1])
            else:
                logger.debug(log)

    console.success(f'Agent {agent_name} built, container [bold red]{container_name}[/] created.')


@agent.command()
@click.option('--file', '-f', type=click.File('rb'), help='Path to Agent yaml definition.', required=True)
@click.option('--organization', '-o', help='Organization name.', required=False, default='')
@click.option('--force/--no-force', default=False)
def build(file: io.FileIO, organization: str = '', force: bool = False) -> None:
    """CLI command to build the agent container from a definition.yaml file.
    Usage : Ostorlab agent build -f path/to/definition.yaml -org organization_name
    """

    if not docker_requirements_checker.is_docker_installed():
        console.error('Docker is not installed.')
        raise click.exceptions.Exit(2)
    if not docker_requirements_checker.is_user_permitted():
        console.error('User does not have permissions to run docker.')
        raise click.exceptions.Exit(2)
    if not docker_requirements_checker.is_docker_working():
        console.error('Error using docker.')
        raise click.exceptions.Exit(2)

    try:
        agent_def = loader.load_agent_yaml(file)
        file.seek(0)
        dockerfile_path = agent_def['docker_file_path']
        docker_build_root = agent_def['docker_build_root']
        agent_name = agent_def['name']
        agent_version = agent_def.get('version', '0.0.0')
        container_name = f'agent_{organization}_{agent_name}:v{agent_version}'
        client = docker.from_env()

        if _image_exists(container_name):
            if force is False:
                console.info(f'{container_name} already exist.')
                raise click.exceptions.Exit(0)
            else:
                console.warning(f'{container_name} already exist, deleting image.')
                client.images.remove(container_name, force=True)

        _build_image(agent_name, container_name, dockerfile_path, docker_build_root, file)

    except errors.BuildError:
        console.error('Error building agent.')
    except validator.SchemaError:
        console.error(
            'Schema is invalid, this should not happen, please report an issue at '
            'https://github.com/Ostorlab/ostorlab/issues.')
    except validator.ValidationError:
        console.error('Definition file does not conform to the provided specification.')



"""Module for the command run inside the group ci_scan.
Example of usage:
- ostorlab --api-key='myKey' ci-scan run --scan-profile=full_scan \
           --break-on-risk-rating=medium --title=test_scan [asset] [options].
"""
import multiprocessing
import click
import time

from ostorlab.cli.ci_scan.ci_scan import ci_scan
from ostorlab.apis import scan_create as scan_create_api
from ostorlab.apis import scan_info as scan_info_api
from ostorlab.apis.runners import authenticated_runner
from ostorlab.cli.ci_scan.run.ci_logger import console_logger, github_logger, logger
from ostorlab.utils import risk_rating

MINUTE = 60
WAIT_MINUTES = 30
SLEEP_CHECKS = 10 # seconds
SCAN_PROGRESS_NOT_STARTED = 'not_started'
SCAN_PROGRESS_DONE = 'done'

CI_LOGGER = {
    'console': console_logger.Logger,
    'github': github_logger.Logger
}

@ci_scan.group()
@click.option('--scan-profile', help='Scan profile to execute.', required=True)
@click.option('--title', help='Scan title.')
@click.option('--break-on-risk-rating', help='Fail if the scan risk rating is higher than the defined value.')
@click.option('--max-wait-minutes', help='Time to wait for the scan results.', required=False, default=WAIT_MINUTES)
@click.option('--log-flavor', help='Type of expected output based on the CI.', required=False, default='console')
@click.pass_context
def run(ctx: click.core.Context, scan_profile: str, title: str,
        break_on_risk_rating: str, max_wait_minutes: int, log_flavor: str) -> None:
    """Start a scan based on a scan profile in the CI.\n"""
    if log_flavor not in CI_LOGGER:
        CI_LOGGER['console']().error(f'log_flavor value {log_flavor} not supported.'
                                     f' Possible options: {CI_LOGGER.keys()}')
    else:
        ci_logger = CI_LOGGER.get(log_flavor)()

    if not ctx.obj.get('api_key'):
        ci_logger.error('API key not not provided.')
        raise click.exceptions.Exit(2)

    if scan_profile in scan_create_api.SCAN_PROFILES:
        ctx.obj['scan_profile'] = scan_create_api.SCAN_PROFILES[scan_profile]
    else:
        ci_logger.error(
            f'Scan profile {scan_profile} not supported. Possible options: {scan_create_api.SCAN_PROFILES.keys()}')

    ctx.obj['title'] = title
    ctx.obj['break_on_risk_rating'] = break_on_risk_rating
    ctx.obj['max_wait_minutes'] = max_wait_minutes
    ctx.obj['ci_logger'] = ci_logger


def apply_break_scan_risk_rating(break_on_risk_rating: str, scan_id: int, max_wait_minutes: int,
                                 runner: authenticated_runner.AuthenticatedAPIRunner, ci_logger: logger.Logger):
    """Wait for the scan to finish and raise an exception if its risk is higher than the defined value."""
    if risk_rating.RiskRating.has_value(break_on_risk_rating.upper()):
        # run the check on a separate process and join it with a timeout. An exception is raised after the timeout.
        check_scan_process = multiprocessing.Process(
            target=_check_scan_periodically,
            args=(runner, scan_id)
        )
        check_scan_process.start()
        check_scan_process.join(int(max_wait_minutes)*MINUTE)

        if check_scan_process.is_alive():
            check_scan_process.kill()
            check_scan_process.join()
            _handle_scan_timeout(runner, scan_id, break_on_risk_rating, ci_logger)

        scan_result = runner.execute(scan_info_api.ScanInfoAPIRequest(scan_id=scan_id))
        if scan_result['data']['scan']['progress'] == 'done':
            scan_risk_rating = scan_result['data']['scan']['riskRating']
            _check_scan_risk_rating(scan_risk_rating, break_on_risk_rating, ci_logger)
    else:
        ci_logger.error(f'Incorrect risk rating value {break_on_risk_rating}. '
                      f'It should be one of {", ".join(risk_rating.RiskRating.values())}')
        raise click.exceptions.Exit(2)


def _check_scan_periodically(runner: authenticated_runner.AuthenticatedAPIRunner, scan_id: int) -> None:
    """Retrieve the scan progress using the API and wait if the progress is different than done"""
    scan_done = False
    while not scan_done:
        scan_result = runner.execute(scan_info_api.ScanInfoAPIRequest(scan_id=scan_id))
        scan_progress = scan_result['data']['scan']['progress']
        if scan_progress == 'done':
            scan_done = True
        time.sleep(SLEEP_CHECKS)


def _check_scan_risk_rating(scan_risk_rating: str, break_on_risk_rating, ci_logger: logger.Logger) -> None:
    """Compare the scan risk and raise an exception if the scan risk is higher than the defined value"""
    if _is_scan_risk_rating_higher(scan_risk_rating, break_on_risk_rating):
        ci_logger.error(f'The scan risk rating is {scan_risk_rating}.')
        raise click.exceptions.Exit(2)
    else:
        ci_logger.info(f'Scan done with risk rating {scan_risk_rating}.')


def _is_scan_risk_rating_higher(scan_risk_rating: str, break_on_risk_rating: str) -> bool:
    """Returns a boolean of the scan risk Comparison with the defined break value"""
    return scan_risk_rating.upper() in risk_rating.RATINGS_ORDER \
           and risk_rating.RATINGS_ORDER[scan_risk_rating.upper()] \
           < risk_rating.RATINGS_ORDER[break_on_risk_rating.upper()]


def _handle_scan_timeout(runner: authenticated_runner.AuthenticatedAPIRunner,
                         scan_id: int, break_on_risk_rating: str, ci_logger: logger.Logger) -> None:
    """when the scan triggers a timeout, we check if the scan has started and if the risk rating is higher
     than the defined break value. In this case we reprot the risk rating and we exit code 2
     otherwise we raise the timeout exception
     """
    scan_result = runner.execute(scan_info_api.ScanInfoAPIRequest(scan_id=scan_id))
    if scan_result['data']['scan']['progress'] != SCAN_PROGRESS_NOT_STARTED:
        scan_risk_rating = scan_result['data']['scan']['riskRating']
        if scan_risk_rating is not None and _is_scan_risk_rating_higher(scan_risk_rating, break_on_risk_rating):
            ci_logger.error(f'The scan risk rating is {scan_risk_rating}.')
            raise click.exceptions.Exit(2)
    raise TimeoutError()



"""Vulnz List command."""
import logging

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name='list')
@click.option('--scan-id', '-s', 'scan_id', help='Id of the scan.', required=True)
@click.pass_context
def list_cli(ctx: click.core.Context, scan_id: int) -> None:
    """CLI command to list vulnerabilities for a scan."""

    runtime_instance = ctx.obj['runtime']
    console.info(f'Fetching vulnerabilities for scan {scan_id}')
    runtime_instance.list_vulnz(scan_id=scan_id)



"""Scan module that handles running a scan using a list of agent keys and a target asset."""

from typing import Optional

import click

from ostorlab.cli.rootcli import rootcli
from ostorlab.runtimes import registry


@rootcli.group()
@click.option('--runtime', type=click.Choice(['local', 'litelocal', 'cloud', 'hybrid']),
              help="""Runtime is in charge of preparing the environment to trigger a scan.\n
                    Specify which runtime to use: \n
                    local: on you local machine\n
                    litelocal: stripped down local runtime\n
                    cloud: on Ostorlab cloud, (requires login)\n
                    hybrid: running partially on Ostorlab cloud and partially on the local machine\n
                   """,
              default='local',
              required=True)
@click.option('--bus-url', help='Bus URL, this flag is restricted to the lite local runtime.', required=False,
              hidden=True)
@click.option('--bus-vhost', help='Bus vhost, this flag is restricted to the lite local runtime.', required=False,
              hidden=True)
@click.option('--bus-management-url', help='Bus management URL, this flag is restricted to the lite local runtime.',
              required=False, hidden=True)
@click.option('--bus-exchange-topic', help='Bus exchange topic, this flag is restricted to the lite local runtime.',
              required=False, hidden=True)
@click.option('--network', help='Docker network to attach service and agents to.',
              required=False, hidden=True)
@click.option('--redis-url', help='Redis URL, this flag is restricted to the lite local runtime.',
              required=False, hidden=True)
@click.option('--scan-id', help='Scan id, this flag is restricted to the lite local runtime..', required=False,
              hidden=True)
@click.pass_context
def scan(ctx: click.core.Context, runtime: str,
         bus_url: Optional[str] = None,
         bus_vhost: Optional[str] = None,
         bus_management_url: Optional[str] = None,
         bus_exchange_topic: Optional[str] = None,
         scan_id: Optional[str] = None,
         network: Optional[str] = None,
         redis_url: Optional[str] = None
         ) -> None:
    """Use scan [subcommand] to list, start or stop a scan.\n
    Examples:\n
        - To show list of scans:\n
        ostorlab scan list\n
        - To stop a scan:\n
        ostorlab scan stop <scan-id>\n
    """
    try:
        runtime_instance = registry.select_runtime(runtime,
                                                   scan_id=scan_id,
                                                   bus_url=bus_url,
                                                   bus_vhost=bus_vhost,
                                                   bus_management_url=bus_management_url,
                                                   bus_exchange_topic=bus_exchange_topic,
                                                   network=network,
                                                   redis_url=redis_url
                                                   )
        ctx.obj['runtime'] = runtime_instance
    except registry.RuntimeNotFoundError as e:
        raise click.ClickException(f'The selected runtime {runtime} is not supported.') from e
    except ValueError as e:
        raise click.ClickException(f'Error initializing runtime {runtime}.') from e



"""Check if requirements for running docker are satisfied."""

import docker
import sys
from docker import errors


def is_docker_installed() -> bool:
    """Checks if docker is installed

    Returns:
        True if docker is installed, else False
    """
    try:
        _ = docker.from_env()
    except errors.DockerException as e:
        if 'ConnectionRefusedError' in str(e):
            return False
    return True


def is_user_permitted() -> bool:
    """Check if the user got permissions to run docker.

    Returns:
        True if user has permission to run docker, else False
    """
    try:
        _ = docker.from_env()
    except errors.DockerException as e:
        if 'PermissionError' in str(e):
            return False
    return True


def is_docker_working() -> bool:
    """Last hope check to see if docker works without being able to give an intelligible recommendation.

    Returns:
        True if user has permission to run docker, else False
    """
    if sys.platform == 'win32':
        import pywintypes  # pylint: disable=import-outside-toplevel
        try:
            client = docker.from_env()
            client.ping()
        except errors.DockerException:
            return False
        except pywintypes.error:
            return False
    else:
        try:
            client = docker.from_env()
            client.ping()
        except errors.DockerException:
            return False
    return True


def is_swarm_initialized() -> bool:
    """Checks if docker swarm is initialized.

    Returns:
        True if docker swarm is initialized, else False
    """
    if is_user_permitted():
        docker_client = docker.from_env()
        if docker_client.swarm.id is None:
            return False
        else:
            return True


def init_swarm() -> None:
    """Initialize docker swarm"""
    if is_user_permitted():
        docker_client = docker.from_env()
        docker_client.swarm.init()



"""Stream logs of a service from a thread."""
import threading
from typing import Generator

import docker.models.services

from ostorlab.cli import console as cli_console

console = cli_console.Console()

COLOR_POOL = [
    'dodger_blue2',
    'deep_sky_blue3',
    'deep_sky_blue2',
    'cyan3',
    'spring_green2',
    'spring_green2',
    'grey37',
    'chartreuse4',
    'cornflower_blue',
    'chartreuse3',
    'steel_blue1',
    'dark_red',
    'plum4',
]


def _stream_log(log_generator: Generator[bytes, None, None], name: str, color: str) -> None:
    """Collect log from log generator and format them using the console API."""
    for line in log_generator:
        console.info(f'[{color} bold]{name}:[/] {line[:-1].decode()}')


class LogStream:
    """Docker service log streamer."""

    def __init__(self):
        self._threads = []
        self._color_map = {}

    def stream(self, service: docker.models.services.Service) -> None:
        """Stream logs of a service without blocking.

        Implementation spawns a dedicated deamon thread.

        Args:
            service: Docker service.
        """
        color = self._select_color(service)
        logs = service.logs(details=False, follow=True, stdout=True, stderr=True)
        t = threading.Thread(target=_stream_log, args=(logs, service.name, color), daemon=False)
        self._threads.append(t)
        t.start()

    def _select_color(self, service):
        """Select color for console output of service."""
        if service.name in self._color_map:
            color = self._color_map[service.name]
        else:
            # To avoid running out of colors, we insert them back.
            color = COLOR_POOL.pop()
            COLOR_POOL.insert(0, color)
        return color



"""Android .APK asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.file.android.apk')
class AndroidApk(asset.Asset):
    """Android .APK  target asset."""
    content: bytes
    path: Optional[str] = None

    def __str__(self):
        return f'Android APK({self.path})'



"""Asset of type .IPA.
This module takes care of preparing a file of type .IPA before calling the create mobile scan API.
"""

import click
import io

from ostorlab.cli.ci_scan.run import run
from ostorlab.cli.ci_scan.run.assets import mobile
from ostorlab.apis import scan_create as scan_create_api


@run.run.command()
@click.argument('file', type=click.File(mode='rb'), required=True)
@click.pass_context
def ios_ipa(ctx: click.core.Context, file: io.FileIO):
    """Create scan for iOS .IPA package file."""
    mobile.run_mobile_scan(ctx, file, scan_create_api.MobileAssetType.IOS)



"""RabbitMQ service in charge of routing Agent messages."""
import binascii
import logging
import os
from typing import Dict

import docker
import tenacity
from docker import errors
from docker import types
from docker.models import services

logger = logging.getLogger(__name__)

MQ_IMAGE = 'rabbitmq:3.9-management'


class LocalRabbitMQ:
    """RabbitMQ service spawned a docker swarm service."""

    def __init__(self,
                 name: str,
                 network: str,
                 exposed_ports: Dict[int, int] = None,
                 image: str = MQ_IMAGE) -> None:
        """Initialize the MQ service parameters.
        Args:
            name: Name of the service.
            network: Network used for the Docker MQ service.
            exposed_ports: The list of MQ service exposed ports
            image: MQ Docker image
        """
        self._name = name
        self._docker_client = docker.from_env()
        # images
        self._mq_image = image
        self._network = network
        self._mq_host = f'mq_{self._name}'
        # service
        self._mq_service = None
        # exposed_port
        self._exposed_ports = exposed_ports

    @property
    def url(self) -> str:
        """URL to connect to the local RabbitMQ instance."""
        return f'amqp://guest:guest@{self._mq_host}:5672/'

    @property
    def vhost(self):
        """Default vhost."""
        return '/'

    @property
    def service(self):
        return self._mq_service

    @property
    def management_url(self) -> str:
        """URL to connect to the management interface of the RabbitMQ instance."""
        return f'http://guest:guest@{self._mq_host}:15672/'

    def start(self) -> None:
        """Start local rabbit mq instance."""
        self._create_network()
        self._mq_service = self._start_mq()

        if self._mq_service is not None and not self._is_service_healthy():
            logger.error('MQ container for service %s is not ready', self._mq_service.id)
            return

    def stop(self):
        for service in self._docker_client.services.list():
            universe = service.attrs['Spec']['Labels'].get('ostorlab.universe')
            if universe is not None and service.name.startswith('mq_') and self._name in universe:
                service.remove()

    def _create_network(self):
        if any(network.name == self._network for network in self._docker_client.networks.list()):
            logger.warning('network already exists.')
        else:
            logger.info('creating private network %s', self._network)
            return self._docker_client.networks.create(
                name=self._network,
                driver='overlay',
                attachable=True,
                labels={'ostorlab.universe': self._name},
                check_duplicate=True
            )

    def _start_mq(self) -> services.Service:
        try:
            logger.info('starting MQ')
            endpoint_spec = types.services.EndpointSpec(mode='vip', ports=self._exposed_ports)
            service_mode = types.services.ServiceMode('replicated', replicas=1)
            return self._docker_client.services.create(
                image=self._mq_image,
                networks=[self._network],
                name=self._mq_host,
                env=[
                    'TASK_ID={{.Task.Slot}}',
                    f'MQ_SERVICE_NAME={self._mq_host}',
                    f'RABBITMQ_ERLANG_COOKIE={binascii.hexlify(os.urandom(10)).decode()}',
                ],
                restart_policy=types.RestartPolicy(condition='any'),
                mode=service_mode,
                labels={'ostorlab.universe': self._name, 'ostorlab.mq': ''},
                endpoint_spec=endpoint_spec)
        except docker.errors.APIError as e:
            error_message = f'MQ service could not be started. Reason: {e}.'
            logger.error(error_message)
            return

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=12),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.outcome.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _is_service_healthy(self) -> bool:
        logger.info('checking service %s', self._mq_service.name)
        return self.is_healthy

    @property
    def is_healthy(self) -> bool:
        try:
            return self._mq_service is not None and \
                   len([task for task in self._mq_service.tasks() if task['Status']['State'] == 'running']) == 1
        except errors.DockerException:
            return False



"""Asset of type Domain Name."""
import logging

import click

from ostorlab.assets import domain_name
from ostorlab.cli.scan.run import run

logger = logging.getLogger(__name__)


@run.run.command(name='domain-name')
@click.argument('name', required=True)
@click.pass_context
def domain_name_cli(ctx: click.core.Context, name: str) -> None:
    """Run scan for Domain Name asset."""
    runtime = ctx.obj['runtime']
    asset = domain_name.DomainName(name=name)
    logger.debug('scanning asset %s', asset)
    runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], asset=asset)
